# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/price.ipynb.

# %% auto 0
__all__ = ['Price']

# %% ../src/price.ipynb 3
from dataclasses import dataclass
from .token import Token

# %% ../src/price.ipynb 4
@dataclass(frozen=True)
class Price:
    """Data class for Token Price

    based on:
    https://github.com/velodrome-finance/oracle/blob/main/contracts/VeloOracle.sol
    """

    token: Token
    price: float

    @property
    def pretty_price(self) -> float: return round(self.price, 5)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/token.ipynb.

# %% auto 0
__all__ = ['Token']

# %% ../src/token.ipynb 2
from typing import Tuple
from dataclasses import dataclass
from .helpers import normalize_address

# %% ../src/token.ipynb 4
@dataclass(frozen=True)
class Token:
    """Data class for Token
    based on: https://github.com/velodrome-finance/sugar/blob/v2/contracts/LpSugar.vy#L17
    """

    token_address: str
    symbol: str
    decimals: int
    listed: bool
    wrapped_token_address: str = None

    @property
    def is_native(self) -> bool: return self.wrapped_token_address is not None

    def value_from_bigint(self, value: float) -> float: return value / 10**self.decimals

    @classmethod
    def make_native_token(cls, symbol: str, wrapped_address: str, decimals: int) -> "Token":
        return Token(token_address=symbol, symbol=symbol, decimals=decimals, listed=True, wrapped_token_address=wrapped_address)

    @classmethod
    def from_tuple(cls, t: Tuple) -> "Token":
        (token_address, symbol, decimals, _, listed) = t
        return Token(
            token_address=normalize_address(token_address),
            symbol=symbol,
            decimals=decimals,
            listed=listed,
        )



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/sugar/',
                'doc_host': 'https://velodrome-finance.github.io',
                'git_url': 'https://github.com/velodrome-finance/sugar-sdk/',
                'lib_path': 'sugar'},
  'syms': { 'sugar.abi': { 'sugar.abi.download_contract_abi': ('abi.html#download_contract_abi', 'sugar/abi.py'),
                           'sugar.abi.get_abi': ('abi.html#get_abi', 'sugar/abi.py')},
            'sugar.chains': { 'sugar.chains.AsyncBaseChain': ('chains.html#asyncbasechain', 'sugar/chains.py'),
                              'sugar.chains.AsyncBaseChain.__init__': ('chains.html#asyncbasechain.__init__', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain': ('chains.html#asyncchain', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.__aenter__': ('chains.html#asyncchain.__aenter__', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.__aexit__': ('chains.html#asyncchain.__aexit__', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain._get_prices': ('chains.html#asyncchain._get_prices', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain._get_quotes_for_paths': ( 'chains.html#asyncchain._get_quotes_for_paths',
                                                                                 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.apaginate': ('chains.html#asyncchain.apaginate', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.check_token_allowance': ( 'chains.html#asyncchain.check_token_allowance',
                                                                                 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.deposit': ('chains.html#asyncchain.deposit', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_all_tokens': ('chains.html#asyncchain.get_all_tokens', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_latest_pool_epochs': ( 'chains.html#asyncchain.get_latest_pool_epochs',
                                                                                  'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_pool_by_address': ( 'chains.html#asyncchain.get_pool_by_address',
                                                                               'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_pool_epochs': ('chains.html#asyncchain.get_pool_epochs', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_pools': ('chains.html#asyncchain.get_pools', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_pools_for_swaps': ( 'chains.html#asyncchain.get_pools_for_swaps',
                                                                               'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_prices': ('chains.html#asyncchain.get_prices', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_quote': ('chains.html#asyncchain.get_quote', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.get_raw_pools': ('chains.html#asyncchain.get_raw_pools', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.set_token_allowance': ( 'chains.html#asyncchain.set_token_allowance',
                                                                               'sugar/chains.py'),
                              'sugar.chains.AsyncChain.sign_and_send_tx': ('chains.html#asyncchain.sign_and_send_tx', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.swap': ('chains.html#asyncchain.swap', 'sugar/chains.py'),
                              'sugar.chains.AsyncChain.swap_from_quote': ('chains.html#asyncchain.swap_from_quote', 'sugar/chains.py'),
                              'sugar.chains.AsyncOPChain': ('chains.html#asyncopchain', 'sugar/chains.py'),
                              'sugar.chains.AsyncOPChain.__init__': ('chains.html#asyncopchain.__init__', 'sugar/chains.py'),
                              'sugar.chains.BaseChain': ('chains.html#basechain', 'sugar/chains.py'),
                              'sugar.chains.BaseChain.__init__': ('chains.html#basechain.__init__', 'sugar/chains.py'),
                              'sugar.chains.BaseChainCommon': ('chains.html#basechaincommon', 'sugar/chains.py'),
                              'sugar.chains.Chain': ('chains.html#chain', 'sugar/chains.py'),
                              'sugar.chains.Chain.__enter__': ('chains.html#chain.__enter__', 'sugar/chains.py'),
                              'sugar.chains.Chain.__exit__': ('chains.html#chain.__exit__', 'sugar/chains.py'),
                              'sugar.chains.Chain._get_prices': ('chains.html#chain._get_prices', 'sugar/chains.py'),
                              'sugar.chains.Chain._get_quotes_for_paths': ('chains.html#chain._get_quotes_for_paths', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_all_tokens': ('chains.html#chain.get_all_tokens', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_latest_pool_epochs': ('chains.html#chain.get_latest_pool_epochs', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_pool_by_address': ('chains.html#chain.get_pool_by_address', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_pool_epochs': ('chains.html#chain.get_pool_epochs', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_pools': ('chains.html#chain.get_pools', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_pools_for_swaps': ('chains.html#chain.get_pools_for_swaps', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_prices': ('chains.html#chain.get_prices', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_quote': ('chains.html#chain.get_quote', 'sugar/chains.py'),
                              'sugar.chains.Chain.get_raw_pools': ('chains.html#chain.get_raw_pools', 'sugar/chains.py'),
                              'sugar.chains.Chain.paginate': ('chains.html#chain.paginate', 'sugar/chains.py'),
                              'sugar.chains.Chain.set_token_allowance': ('chains.html#chain.set_token_allowance', 'sugar/chains.py'),
                              'sugar.chains.Chain.sign_and_send_tx': ('chains.html#chain.sign_and_send_tx', 'sugar/chains.py'),
                              'sugar.chains.Chain.swap': ('chains.html#chain.swap', 'sugar/chains.py'),
                              'sugar.chains.Chain.swap_from_quote': ('chains.html#chain.swap_from_quote', 'sugar/chains.py'),
                              'sugar.chains.CommonChain': ('chains.html#commonchain', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.__init__': ('chains.html#commonchain.__init__', 'sugar/chains.py'),
                              'sugar.chains.CommonChain._prepare_prices': ('chains.html#commonchain._prepare_prices', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.account': ('chains.html#commonchain.account', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.filter_pools_for_swap': ( 'chains.html#commonchain.filter_pools_for_swap',
                                                                                  'sugar/chains.py'),
                              'sugar.chains.CommonChain.get_paths_for_quote': ( 'chains.html#commonchain.get_paths_for_quote',
                                                                                'sugar/chains.py'),
                              'sugar.chains.CommonChain.get_pool_paginator': ( 'chains.html#commonchain.get_pool_paginator',
                                                                               'sugar/chains.py'),
                              'sugar.chains.CommonChain.paths_to_pools': ('chains.html#commonchain.paths_to_pools', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_pool_epochs': ( 'chains.html#commonchain.prepare_pool_epochs',
                                                                                'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_pools': ('chains.html#commonchain.prepare_pools', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_pools_for_swap': ( 'chains.html#commonchain.prepare_pools_for_swap',
                                                                                   'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_price_batcher': ( 'chains.html#commonchain.prepare_price_batcher',
                                                                                  'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_prices': ('chains.html#commonchain.prepare_prices', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_quote_batch': ( 'chains.html#commonchain.prepare_quote_batch',
                                                                                'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_quotes': ('chains.html#commonchain.prepare_quotes', 'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_set_token_allowance_contract': ( 'chains.html#commonchain.prepare_set_token_allowance_contract',
                                                                                                 'sugar/chains.py'),
                              'sugar.chains.CommonChain.prepare_tokens': ('chains.html#commonchain.prepare_tokens', 'sugar/chains.py'),
                              'sugar.chains.OPChain': ('chains.html#opchain', 'sugar/chains.py'),
                              'sugar.chains.OPChain.__init__': ('chains.html#opchain.__init__', 'sugar/chains.py'),
                              'sugar.chains.OPChainCommon': ('chains.html#opchaincommon', 'sugar/chains.py'),
                              'sugar.chains.require_async_context': ('chains.html#require_async_context', 'sugar/chains.py'),
                              'sugar.chains.require_context': ('chains.html#require_context', 'sugar/chains.py'),
                              'sugar.chains.safe_format_batched_response': ('chains.html#safe_format_batched_response', 'sugar/chains.py')},
            'sugar.config': { 'sugar.config.ChainSettings': ('config.html#chainsettings', 'sugar/config.py'),
                              'sugar.config.ChainSettings.__repr__': ('config.html#chainsettings.__repr__', 'sugar/config.py'),
                              'sugar.config.ChainSettings.__str__': ('config.html#chainsettings.__str__', 'sugar/config.py'),
                              'sugar.config.get_env': ('config.html#get_env', 'sugar/config.py'),
                              'sugar.config.make_base_chain_settings': ('config.html#make_base_chain_settings', 'sugar/config.py'),
                              'sugar.config.make_op_chain_settings': ('config.html#make_op_chain_settings', 'sugar/config.py'),
                              'sugar.config.make_settings': ('config.html#make_settings', 'sugar/config.py'),
                              'sugar.config.validate_settings': ('config.html#validate_settings', 'sugar/config.py')},
            'sugar.deposit': {'sugar.deposit.Deposit': ('deposit.html#deposit', 'sugar/deposit.py')},
            'sugar.helpers': { 'sugar.helpers.Pair': ('helpers.html#pair', 'sugar/helpers.py'),
                               'sugar.helpers.Timer': ('helpers.html#timer', 'sugar/helpers.py'),
                               'sugar.helpers.Timer.__aenter__': ('helpers.html#timer.__aenter__', 'sugar/helpers.py'),
                               'sugar.helpers.Timer.__aexit__': ('helpers.html#timer.__aexit__', 'sugar/helpers.py'),
                               'sugar.helpers.Timer.__enter__': ('helpers.html#timer.__enter__', 'sugar/helpers.py'),
                               'sugar.helpers.Timer.__exit__': ('helpers.html#timer.__exit__', 'sugar/helpers.py'),
                               'sugar.helpers.Timer.__init__': ('helpers.html#timer.__init__', 'sugar/helpers.py'),
                               'sugar.helpers.amount_to_k_string': ('helpers.html#amount_to_k_string', 'sugar/helpers.py'),
                               'sugar.helpers.amount_to_m_string': ('helpers.html#amount_to_m_string', 'sugar/helpers.py'),
                               'sugar.helpers.apply_slippage': ('helpers.html#apply_slippage', 'sugar/helpers.py'),
                               'sugar.helpers.atime_it': ('helpers.html#atime_it', 'sugar/helpers.py'),
                               'sugar.helpers.chunk': ('helpers.html#chunk', 'sugar/helpers.py'),
                               'sugar.helpers.find_all_paths': ('helpers.html#find_all_paths', 'sugar/helpers.py'),
                               'sugar.helpers.float_to_uint256': ('helpers.html#float_to_uint256', 'sugar/helpers.py'),
                               'sugar.helpers.format_currency': ('helpers.html#format_currency', 'sugar/helpers.py'),
                               'sugar.helpers.format_percentage': ('helpers.html#format_percentage', 'sugar/helpers.py'),
                               'sugar.helpers.get_future_timestamp': ('helpers.html#get_future_timestamp', 'sugar/helpers.py'),
                               'sugar.helpers.normalize_address': ('helpers.html#normalize_address', 'sugar/helpers.py'),
                               'sugar.helpers.time_it': ('helpers.html#time_it', 'sugar/helpers.py')},
            'sugar.pool': { 'sugar.pool.Amount': ('pool.html#amount', 'sugar/pool.py'),
                            'sugar.pool.Amount.amount_in_stable': ('pool.html#amount.amount_in_stable', 'sugar/pool.py'),
                            'sugar.pool.Amount.build': ('pool.html#amount.build', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool': ('pool.html#liquiditypool', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.apr': ('pool.html#liquiditypool.apr', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.from_tuple': ('pool.html#liquiditypool.from_tuple', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.pool_fee_percentage': ( 'pool.html#liquiditypool.pool_fee_percentage',
                                                                              'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.token0_volume': ('pool.html#liquiditypool.token0_volume', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.token1_volume': ('pool.html#liquiditypool.token1_volume', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.total_fees': ('pool.html#liquiditypool.total_fees', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.tvl': ('pool.html#liquiditypool.tvl', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.volume': ('pool.html#liquiditypool.volume', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPool.volume_pct': ('pool.html#liquiditypool.volume_pct', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolEpoch': ('pool.html#liquiditypoolepoch', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolEpoch.epoch_date': ('pool.html#liquiditypoolepoch.epoch_date', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolEpoch.from_tuple': ('pool.html#liquiditypoolepoch.from_tuple', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolEpoch.total_fees': ('pool.html#liquiditypoolepoch.total_fees', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolEpoch.total_incentives': ( 'pool.html#liquiditypoolepoch.total_incentives',
                                                                                'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolForSwap': ('pool.html#liquiditypoolforswap', 'sugar/pool.py'),
                            'sugar.pool.LiquidityPoolForSwap.from_tuple': ('pool.html#liquiditypoolforswap.from_tuple', 'sugar/pool.py'),
                            'sugar.pool.Price': ('pool.html#price', 'sugar/pool.py'),
                            'sugar.pool.Price.pretty_price': ('pool.html#price.pretty_price', 'sugar/pool.py'),
                            'sugar.pool.symbol': ('pool.html#symbol', 'sugar/pool.py')},
            'sugar.price': { 'sugar.price.Price': ('price.html#price', 'sugar/price.py'),
                             'sugar.price.Price.pretty_price': ('price.html#price.pretty_price', 'sugar/price.py')},
            'sugar.quote': { 'sugar.quote.PreparedRoute': ('quote.html#preparedroute', 'sugar/quote.py'),
                             'sugar.quote.PreparedRoute.encoded': ('quote.html#preparedroute.encoded', 'sugar/quote.py'),
                             'sugar.quote.Quote': ('quote.html#quote', 'sugar/quote.py'),
                             'sugar.quote.Quote.amount_in': ('quote.html#quote.amount_in', 'sugar/quote.py'),
                             'sugar.quote.Quote.from_token': ('quote.html#quote.from_token', 'sugar/quote.py'),
                             'sugar.quote.Quote.path': ('quote.html#quote.path', 'sugar/quote.py'),
                             'sugar.quote.Quote.to_token': ('quote.html#quote.to_token', 'sugar/quote.py'),
                             'sugar.quote.QuoteInput': ('quote.html#quoteinput', 'sugar/quote.py'),
                             'sugar.quote.QuoteInput.from_tuple': ('quote.html#quoteinput.from_tuple', 'sugar/quote.py'),
                             'sugar.quote.QuoteInput.route': ('quote.html#quoteinput.route', 'sugar/quote.py'),
                             'sugar.quote.QuoteInput.to_tuple': ('quote.html#quoteinput.to_tuple', 'sugar/quote.py'),
                             'sugar.quote.pack_path': ('quote.html#pack_path', 'sugar/quote.py')},
            'sugar.swap': { 'sugar.swap.CommandType': ('swap.html#commandtype', 'sugar/swap.py'),
                            'sugar.swap.RoutePlanner': ('swap.html#routeplanner', 'sugar/swap.py'),
                            'sugar.swap.RoutePlanner.__init__': ('swap.html#routeplanner.__init__', 'sugar/swap.py'),
                            'sugar.swap.RoutePlanner.add_command': ('swap.html#routeplanner.add_command', 'sugar/swap.py'),
                            'sugar.swap.RoutePlanner.get_encoded_commands': ( 'swap.html#routeplanner.get_encoded_commands',
                                                                              'sugar/swap.py'),
                            'sugar.swap.RoutePlanner.get_encoded_inputs': ('swap.html#routeplanner.get_encoded_inputs', 'sugar/swap.py'),
                            'sugar.swap.RoutePlanner.get_pretty_encoded_inputs': ( 'swap.html#routeplanner.get_pretty_encoded_inputs',
                                                                                   'sugar/swap.py'),
                            'sugar.swap.setup_planner': ('swap.html#setup_planner', 'sugar/swap.py')},
            'sugar.token': { 'sugar.token.Token': ('token.html#token', 'sugar/token.py'),
                             'sugar.token.Token.from_tuple': ('token.html#token.from_tuple', 'sugar/token.py'),
                             'sugar.token.Token.is_native': ('token.html#token.is_native', 'sugar/token.py'),
                             'sugar.token.Token.make_native_token': ('token.html#token.make_native_token', 'sugar/token.py'),
                             'sugar.token.Token.value_from_bigint': ('token.html#token.value_from_bigint', 'sugar/token.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/pool.ipynb.

# %% auto 0
__all__ = ['Price', 'Amount', 'symbol', 'LiquidityPoolForSwap', 'LiquidityPool', 'LiquidityPoolEpoch']

# %% ../src/pool.ipynb 3
from dataclasses import dataclass
from datetime import datetime
from typing import Tuple, Dict, Optional, List
from .token import Token
from .helpers import normalize_address

# %% ../src/pool.ipynb 4
@dataclass(frozen=True)
class Price:
    """Data class for Token Price

    based on:
    https://github.com/velodrome-finance/oracle/blob/main/contracts/VeloOracle.sol
    """

    token: Token
    price: float

    @property
    def pretty_price(self) -> float: return round(self.price, 5)

# %% ../src/pool.ipynb 5
@dataclass(frozen=True)
class Amount:
    token: Token
    amount: float
    price: "Price"

    @classmethod
    def build(cls, address: str, amount: float, tokens: Dict[str, Token], prices: Dict[str, "Price"]) -> "Amount":
        address = normalize_address(address)
        if address not in tokens or address not in prices: return None
        token = tokens[address]
        return Amount(token=token, amount=token.value_from_bigint(amount), price=prices[address])

    @property
    def amount_in_stable(self) -> float: return self.amount * self.price.price


# %% ../src/pool.ipynb 6
def symbol(token0: Token, token1: Token, pool_type: int) -> str:
    return f"CL{pool_type}-{token0.symbol}/{token1.symbol}" if pool_type > 0  else f"{'s' if pool_type == 0 else 'v'}AMM-{token0.symbol}/{token1.symbol}"

@dataclass(frozen=True)
class LiquidityPoolForSwap:
    """A more compact pool representation used for swaps"""

    lp: str
    type: int
    token0_address: str
    token1_address: str
    is_stable: bool
    # concentrated liquidity pools
    is_cl: bool

    @classmethod
    def from_tuple(cls, t: Tuple) -> "LiquidityPoolForSwap":
        token0, token1, pool_type = normalize_address(t[2]), normalize_address(t[3]), t[1]
        return LiquidityPoolForSwap(
            lp=normalize_address(t[0]),
            type=pool_type,
            token0_address=token0,
            token1_address=token1,
            is_stable=pool_type == 0,
            is_cl=pool_type > 0
        )

        


@dataclass(frozen=True)
class LiquidityPool:
    """Data class for Liquidity Pool

    based on:
    https://github.com/velodrome-finance/sugar/blob/v2/contracts/LpSugar.vy#L31
    """

    lp: str
    factory: str
    symbol: str
    type: int
    is_stable: bool
    # concentrated liquidity pools
    is_cl: bool
    total_supply: float
    decimals: int
    token0: Token
    reserve0: Amount
    token1: Token
    reserve1: Amount
    token0_fees: Amount
    token1_fees: Amount
    pool_fee: float
    gauge_total_supply: float
    emissions: Amount
    emissions_token: Token
    weekly_emissions: Amount
    nfpm: str
    alm: str

    @property
    def tvl(self) -> float: return self.reserve0.amount_in_stable + self.reserve1.amount_in_stable

    @property
    def total_fees(self) -> float:
        result = 0

        if self.token0_fees: result += self.token0_fees.amount_in_stable
        if self.token1_fees: result += self.token1_fees.amount_in_stable

        return result

    @property
    def pool_fee_percentage(self) -> float: return self.pool_fee / 100

    @property
    def volume_pct(self) -> float: return 100 / self.pool_fee_percentage

    @property
    def volume(self) -> Amount:
        t0 = self.token0_fees.amount_in_stable if self.token0_fees else 0
        t1 = self.token1_fees.amount_in_stable if self.token1_fees else 0
        return self.volume_pct * (t0 + t1)

    @property
    def token0_volume(self) -> float: return self.token0_fees.amount * self.volume_pct if self.token0_fees else 0

    @property
    def token1_volume(self) -> float: return self.token1_fees.amount * self.volume_pct if self.token1_fees else 0

    @property
    def apr(self) -> float:
        day_seconds, tvl = 24 * 60 * 60, self.tvl
        reward_value = self.emissions.amount_in_stable if self.emissions else 0
        reward = reward_value * day_seconds
        staked_pct = (
            100 * self.gauge_total_supply / self.total_supply if self.total_supply != 0 else 0
        )
        staked_tvl = tvl * staked_pct / 100
        return (reward / staked_tvl) * (100 * 365) if staked_tvl != 0 else 0

    @classmethod
    def from_tuple(
        cls, t: Tuple, tokens: Dict[str, Token], prices: Dict[str, Price]
    ) -> Optional["LiquidityPool"]:
        token0, token1, pool_type = normalize_address(t[7]), normalize_address(t[10]), t[4]
        token0_fees, token1_fees = t[23], t[24]
        emissions_token = normalize_address(t[20])
        emissions = t[19]

        seconds_in_a_week = 7 * 24 * 60 * 60

        # Sugar.all returns a tuple with the following structure:
        # { "name": "lp", "type": "address" },          <== 0
        # { "name": "symbol", "type": "string" },       <== 1
        # { "name": "decimals", "type": "uint8" },      <== 2
        # { "name": "liquidity", "type": "uint256" },   <== 3
        # { "name": "type", "type": "int24" },          <== 4
        # { "name": "tick", "type": "int24" },          <== 5
        # { "name": "sqrt_ratio", "type": "uint160" },  <== 6
        # { "name": "token0", "type": "address" },      <== 7
        # { "name": "reserve0", "type": "uint256" },    <== 8
        # { "name": "staked0", "type": "uint256" },     <== 9
        # { "name": "token1", "type": "address" },      <== 10
        # { "name": "reserve1", "type": "uint256" },    <== 11
        # { "name": "staked1", "type": "uint256" },     <== 12
        # { "name": "gauge", "type": "address" },        <== 13
        # { "name": "gauge_liquidity", "type": "uint256" },  <== 14
        # { "name": "gauge_alive", "type": "bool" },        <== 15
        # { "name": "fee", "type": "address" },             <== 16
        # { "name": "bribe", "type": "address" },           <== 17
        # { "name": "factory", "type": "address" },         <== 18
        # { "name": "emissions", "type": "uint256" },       <== 19
        # { "name": "emissions_token", "type": "address" },  <== 20
        # { "name": "pool_fee", "type": "uint256" },        <== 21
        # { "name": "unstaked_fee", "type": "uint256" },    <== 22
        # { "name": "token0_fees", "type": "uint256" },     <== 23
        # { "name": "token1_fees", "type": "uint256" },    <== 24
        # { "name": "nfpm", "type": "address" },           <== 25
        # { "name": "alm", "type": "address" }             <== 26

        token0, token1 = tokens.get(token0), tokens.get(token1)
        if not token0 or not token1: return None

        return LiquidityPool(
            lp=normalize_address(t[0]),
            factory=normalize_address(t[18]),
            symbol=symbol(token0, token1, pool_type),
            type=pool_type,
            # stable pools have type set to 0
            is_stable=pool_type == 0,
            is_cl=pool_type > 0,
            total_supply=t[3],
            decimals=t[2],
            token0=token0,
            reserve0=Amount.build(token0.token_address, t[8], tokens, prices),
            token1=token1,
            reserve1=Amount.build(token1.token_address, t[11], tokens, prices),
            token0_fees=Amount.build(token0.token_address, token0_fees, tokens, prices),
            token1_fees=Amount.build(token1.token_address, token1_fees, tokens, prices),
            pool_fee=t[21],
            gauge_total_supply=t[14],
            emissions_token=tokens.get(emissions_token),
            emissions=Amount.build(emissions_token, emissions, tokens, prices),
            weekly_emissions=Amount.build(emissions_token, emissions * seconds_in_a_week, tokens, prices),
            nfpm=normalize_address(t[25]),
            alm=normalize_address(t[26]),
        )

# %% ../src/pool.ipynb 7
@dataclass(frozen=True)
class LiquidityPoolEpoch:
    ts: int
    lp: str
    pool: LiquidityPool
    votes: int
    emissions: int
    incentives: List[Amount]
    fees: List[Amount]

    @classmethod
    def from_tuple(cls, t: Tuple, pools: Dict[str, LiquidityPool], tokens: Dict[str, Token], prices: Dict[str, Price]) -> "LiquidityPoolEpoch":
        ts, lp, votes, emissions, incentives, fees = t[0], normalize_address(t[1]), t[2], t[3], t[4], t[5]
        def build_amount(token_address: str, amount: int) -> Amount:
            return Amount.build(normalize_address(token_address), amount, tokens, prices)

        return LiquidityPoolEpoch(ts=ts, lp=lp, pool=pools.get(lp), votes=votes, emissions=emissions,
            incentives=[build_amount(i[0], i[1]) for i in incentives],
            fees=[build_amount(f[0], f[1]) for f in fees]
        )

    @property
    def total_fees(self) -> float:
        return sum([fee.amount_in_stable for fee in self.fees]) if self.fees else 0
    @property
    def total_incentives(self) -> float:
        return sum([incentive.amount_in_stable for incentive in self.incentives]) if self.incentives else 0

    @property
    def epoch_date(self) -> datetime:
        return datetime.fromtimestamp(self.ts)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/swap.ipynb.

# %% auto 0
__all__ = ['ABI_DEFINITION', 'CONTRACT_BALANCE_FOR_V3_SWAPS', 'CommandType', 'RoutePlanner', 'setup_planner']

# %% ../src/swap.ipynb 3
from .quote import Quote, pack_path
from .helpers import apply_slippage
from .pool import LiquidityPoolForSwap
from enum import IntEnum
from typing import List, Any, Tuple
from eth_abi import encode

# %% ../src/swap.ipynb 4
class CommandType(IntEnum):
    V3_SWAP_EXACT_IN = 0x00
    V3_SWAP_EXACT_OUT = 0x01
    SWEEP = 0x04
    V2_SWAP_EXACT_IN = 0x08
    V2_SWAP_EXACT_OUT = 0x09
    WRAP_ETH = 0x0b
    UNWRAP_WETH = 0x0c

# Define ABI types for each command
ABI_DEFINITION = {
    CommandType.V3_SWAP_EXACT_IN: [
        "address",
        "uint256",
        "uint256",
        "bytes",
        "bool"
    ],
    CommandType.V2_SWAP_EXACT_IN: [
        "address",
        "uint256",
        "uint256",
        "(address,address,bool)[]",
        "bool"
    ],
    CommandType.V2_SWAP_EXACT_OUT: [
        "address",
        "uint256",
        "uint256",
        "(address,address,bool)[]",
        "bool"
    ],
    CommandType.V3_SWAP_EXACT_OUT: [
        "address",
        "uint256",
        "uint256",
        "bytes",
        "bool"
    ],
    CommandType.WRAP_ETH: [
        "address",
        "uint256"
    ],
    CommandType.UNWRAP_WETH: [
        "address",
        "uint256"
    ],
    CommandType.SWEEP: [
        "address",
        "address",
        "uint256"
    ]
}

class RoutePlanner:
    def __init__(self):
        """Initialize a new RoutePlanner"""
        self.commands = "0x"
        self.inputs: List[bytes] = []

    def add_command(self, command_type: CommandType, parameters: List[Any]) -> None:
        """
        Add a command to the route planner
        
        Args:
            command_type: Type of command to add
            parameters: Parameters for the command
        """
        # Get the ABI definition for this command
        abi_types = ABI_DEFINITION[command_type]
        self.inputs.append(encode(abi_types, parameters))
        # Add command byte to commands
        command_hex = format(command_type, '02x')
        self.commands = self.commands + command_hex

    def get_encoded_commands(self) -> str: return self.commands
    
    def get_encoded_inputs(self) -> List[bytes]: return self.inputs

    # using this for testing
    def get_pretty_encoded_inputs(self) -> List[str]: return list(map(lambda i: "0x" + i.hex(), self.get_encoded_inputs())) 

# %% ../src/swap.ipynb 5
# Constants
CONTRACT_BALANCE_FOR_V3_SWAPS = int("0x8000000000000000000000000000000000000000000000000000000000000000", 16)

def setup_planner(quote: Quote, slippage: float, account: str, router_address) -> RoutePlanner:
    """Setup route planner with the given quote and chain"""
    
    route_planner = RoutePlanner()
    min_amount_out = apply_slippage(quote.amount_out, slippage)

    # By default money comes from contract
    tokens_come_from_contract = False
    
    # Handle wrapped native token if needed
    if quote.from_token.wrapped_token_address:
        # When trading from native token, wrap token first
        route_planner.add_command(CommandType.WRAP_ETH, [router_address, quote.amount_in])
        tokens_come_from_contract = True
    
    # Group nodes by pool type (v2 or v3)
    grouped_nodes: List[List[Tuple[LiquidityPoolForSwap, bool]]] = []
    
    # helpers for getting from and to token addresses
    def from_token_address(node: Tuple[LiquidityPoolForSwap, bool]) -> str:
        pool, reversed = node
        return pool.token0_address if not reversed else pool.token1_address
    def to_token_address(node: Tuple[LiquidityPoolForSwap, bool]) -> str:
        pool, reversed = node
        return pool.token1_address if not reversed else pool.token0_address


    for node in quote.path:
        if not grouped_nodes: grouped_nodes.append([node])
        elif node[0].type < 1:
            # Current node is a v2 pool
            if float(grouped_nodes[-1][0][0].type) < 1: grouped_nodes[-1].append(node)
            else: grouped_nodes.append([node])
        else:
            # Current node is a v3 pool
            if grouped_nodes[-1][0][0].type >= 1: grouped_nodes[-1].append(node)
            else: grouped_nodes.append([node])
    
    if len(grouped_nodes) == 1:
        # All nodes belong to the same pool type
        nodes = grouped_nodes[0]
        is_v2_pool = float(nodes[0][0].type) < 1
        
        route_planner.add_command(
            CommandType.V2_SWAP_EXACT_IN if is_v2_pool else CommandType.V3_SWAP_EXACT_IN,
            [
                # Where should money go?
                router_address if quote.to_token.wrapped_token_address else account,
                quote.amount_in,
                min_amount_out,
                [
                    # from, to, stable
                    (from_token_address(n), to_token_address(n), n[0].is_stable) for n in nodes
                ] if is_v2_pool else pack_path(nodes).encoded,
                not tokens_come_from_contract,
            ]
        )
    else:
        # Mixed v2 and v3 pools
        first_batch = grouped_nodes[0]
        last_batch = grouped_nodes[-1]
        rest = grouped_nodes[1:-1]
        
        # Handle first batch
        is_first_batch_v2 = not first_batch[0][0].is_cl
        next_batch = rest[0] if rest else last_batch
        
        route_planner.add_command(
            CommandType.V2_SWAP_EXACT_IN if is_first_batch_v2 else CommandType.V3_SWAP_EXACT_IN,
            [
                router_address if is_first_batch_v2 else next_batch[0][0].lp,
                quote.amount_in,
                0,  # No expectations on min amount out for first batch
                [
                    # from, to, stable
                    (from_token_address(n), to_token_address(n), n[0].is_stable) for n in first_batch
                ] if is_first_batch_v2 else pack_path(first_batch).encoded,
                not tokens_come_from_contract,
            ]
        )
        
        # Handle middle batches
        for idx, batch in enumerate(rest):
            is_batch_v2 = not batch[0][0].is_cl
            next_batch = rest[idx + 1] if idx + 1 < len(rest) else last_batch
            
            route_planner.add_command(
                CommandType.V2_SWAP_EXACT_IN if is_batch_v2 else CommandType.V3_SWAP_EXACT_IN,
                [
                    router_address if is_batch_v2 else next_batch[0][0].lp,
                    0 if is_batch_v2 else CONTRACT_BALANCE_FOR_V3_SWAPS,
                    0,  # No expectations for middle batches
                    [
                        (from_token_address(n), to_token_address(n), n[0].is_stable) for n in batch
                    ] if is_batch_v2 else pack_path(batch).encoded,
                    False,  # Money comes from contract
                ]
            )
        
        # # Handle last batch
        is_last_batch_v2 = not last_batch[0][0].is_cl
        
        route_planner.add_command(
            CommandType.V2_SWAP_EXACT_IN if is_last_batch_v2 else CommandType.V3_SWAP_EXACT_IN,
            [
                router_address if quote.to_token.wrapped_token_address else account,
                0 if is_last_batch_v2 else CONTRACT_BALANCE_FOR_V3_SWAPS,
                min_amount_out,
                [
                    (from_token_address(n), to_token_address(n), n[0].is_stable) for n in last_batch
                ] if is_last_batch_v2 else pack_path(last_batch).encoded,
                False,  # Money comes from contract
            ]
        )
    
    # Handle unwrapping WETH if needed
    if quote.to_token.wrapped_token_address: route_planner.add_command(CommandType.UNWRAP_WETH, [account, min_amount_out])
    
    return route_planner



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/deposit.ipynb.

# %% auto 0
__all__ = ['Deposit']

# %% ../src/deposit.ipynb 2
from dataclasses import dataclass
from .pool import LiquidityPool

# %% ../src/deposit.ipynb 3
@dataclass(frozen=True)
class Deposit:
    """Data class for Deposits"""
    # pool to deposit into
    pool: LiquidityPool
    # amount of token0 for deposit
    amount_token0: float



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/chains.ipynb.

# %% auto 0
__all__ = ['original_format_batched_response', 'T', 'safe_format_batched_response', 'require_context', 'require_async_context',
           'CommonChain', 'AsyncChain', 'Chain', 'OPChainCommon', 'AsyncOPChain', 'OPChain', 'BaseChainCommon',
           'AsyncBaseChain', 'BaseChain']

# %% ../src/chains.ipynb 3
import os, asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import wraps, lru_cache
from async_lru import alru_cache
from cachetools import cached, TTLCache
from typing import List, TypeVar, Callable, Optional, Tuple, Dict
from web3 import Web3, HTTPProvider, AsyncWeb3, AsyncHTTPProvider, Account
from web3.eth.async_eth import AsyncContract
from web3.eth import Contract
from web3.manager import RequestManager, RequestBatcher
from .config import ChainSettings, make_op_chain_settings, make_base_chain_settings
from .helpers import normalize_address, MAX_UINT256, float_to_uint256, apply_slippage, get_future_timestamp, ADDRESS_ZERO, chunk, Pair
from .helpers import find_all_paths, time_it, atime_it
from .abi import get_abi
from .token import Token
from .pool import LiquidityPool, LiquidityPoolForSwap, LiquidityPoolEpoch
from .price import Price
from .deposit import Deposit
from .quote import QuoteInput, Quote
from .swap import setup_planner

# %% ../src/chains.ipynb 4
# monkey patching how web3 handles errors in batched requests
# re: https://github.com/ethereum/web3.py/issues/3657
original_format_batched_response = RequestManager._format_batched_response
def safe_format_batched_response(*args):
    try: return original_format_batched_response(*args)
    except Exception as e: return e
RequestManager._format_batched_response = safe_format_batched_response

# %% ../src/chains.ipynb 6
T = TypeVar('T')

def require_context(f: Callable[..., T]) -> Callable[..., T]:
    @wraps(f)
    def wrapper(self: 'CommonChain', *args, **kwargs) -> T:
        if not self._in_context: raise RuntimeError("Chain methods can only be accessed within 'async with' block")
        return f(self, *args, **kwargs)
    return wrapper

def require_async_context(f: Callable[..., T]) -> Callable[..., T]:
    @wraps(f)
    async def wrapper(self: 'CommonChain', *args, **kwargs) -> T:
        if not self._in_context: raise RuntimeError("Chain methods can only be accessed within 'async with' block")
        return await f(self, *args, **kwargs)
    return wrapper


class CommonChain:
    @property
    def account(self) -> Account: return self.web3.eth.account.from_key(os.getenv("SUGAR_PK"))

    pools: Optional[List[LiquidityPool]] = None
    pools_for_swap: Optional[List[LiquidityPoolForSwap]] = None

    def __init__(self, settings: ChainSettings, **kwargs):
        self.settings, self._in_context = settings, False

        if "pools" in kwargs: self.pools = kwargs["pools"]
        if "pools_for_swap" in kwargs: self.pools_for_swap = kwargs["pools_for_swap"] 
    
    def prepare_set_token_allowance_contract(self, token: Token, contract_wrapper):
        ERC20_ABI = [{
            "name": "approve",
            "type": "function",
            "constant": False,
            "inputs": [{"name": "spender", "type": "address"}, {"name": "amount", "type": "uint256"}],
            "outputs": [{"name": "", "type": "bool"}]
        }]
        return contract_wrapper(address=token.wrapped_token_address or token.token_address, abi=ERC20_ABI)
    
    def prepare_tokens(self, tokens: List[Tuple], listed_only: bool) -> List[Token]:
        native = Token.make_native_token(self.settings.native_token_symbol, self.settings.wrapped_native_token_addr, self.settings.native_token_decimals)
        ts = list(map(lambda t: Token.from_tuple(t), tokens))
        return [native] + (list(filter(lambda t: t.listed, ts)) if listed_only else ts)
    
    def _prepare_prices(self, tokens: List[Token], rates: List[int]) -> Dict[str, int]:
        # token_address => normalized rate
        result = {}
        # rates are returned multiplied by eth decimals + the difference in decimals to eth
        # we want them all normalized to 18 decimals
        for cnt, rate in enumerate(rates):
            t, eth_decimals = tokens[cnt], self.settings.native_token_decimals
            if t.decimals == eth_decimals: nr = rate
            elif t.decimals < eth_decimals: nr = rate // (10 ** (eth_decimals - t.decimals))
            else: nr = rate * (10 ** (t.decimals - eth_decimals))
            result[t.token_address] = nr
        return result

    def prepare_prices(self, tokens: List[Token], prices: List[int]) -> List[Price]:
        """Get prices for tokens in target stable token"""
        eth_decimals = self.settings.native_token_decimals
        # all rates in EHT: token => rate
        rates_in_eth = self._prepare_prices(tokens, prices)
        eth_rate, usd_rate = rates_in_eth[self.settings.native_token_symbol], rates_in_eth[self.settings.stable_token_addr]
        # this gives us the price of 1 eth in usd with 18 decimals precision
        eth_usd_price = (eth_rate * 10 ** eth_decimals) // usd_rate
        # finally convert to prices in terms of stable
        return [Price(token=t, price=(rates_in_eth[t.token_address] * eth_usd_price // 10 ** eth_decimals) / 10 ** eth_decimals) for t in tokens]
    
    def prepare_pools(self, pools: List[Tuple], tokens: List[Token], prices: List[Price]) -> List[LiquidityPool]:
        tokens, prices = {t.token_address: t for t in tokens}, {price.token.token_address: price for price in prices}
        return list(filter(lambda p: p is not None, map(lambda p: LiquidityPool.from_tuple(p, tokens, prices), pools)))
    
    def prepare_pools_for_swap(self, pools: List[Tuple]) -> List[LiquidityPoolForSwap]:
        return list(map(lambda p: LiquidityPoolForSwap.from_tuple(p), pools))

    def prepare_pool_epochs(self, epochs: List[Tuple], pools: List[LiquidityPool], tokens: List[Token], prices: List[Price]) -> List[LiquidityPoolEpoch]:
        tokens, prices, pools = {t.token_address: t for t in tokens}, {price.token.token_address: price for price in prices}, {p.lp: p for p in pools}
        return list(map(lambda p: LiquidityPoolEpoch.from_tuple(p, pools, tokens, prices), epochs))
    
    def filter_pools_for_swap(self, pools: List[LiquidityPoolForSwap], from_token: Token, to_token: Token) -> List[LiquidityPoolForSwap]:
        match_tokens = set(self.settings.connector_tokens_addrs + [from_token.token_address, to_token.token_address])
        return list(filter(lambda p: p.token0_address in match_tokens or p.token1_address in match_tokens, pools))
    
    def paths_to_pools(self, pools: List[LiquidityPoolForSwap], paths: List[List[Tuple]]) -> List[LiquidityPoolForSwap]:
        pools_dict = {p.lp: p for p in pools}
        return [list(map(lambda p: pools_dict[p[2]], path)) for path in paths]

    def prepare_quote_batch(self, from_token: Token, to_token: Token, batcher: RequestBatcher, pools: List[List[LiquidityPoolForSwap]], amount_in: int, paths: List[List[Tuple]]):
        inputs = []
        for i, path in enumerate(paths):
            p = [(p, p.token0_address != path[i][0]) for i, p in enumerate(pools[i])]
            q = QuoteInput(from_token=from_token, to_token=to_token, amount_in=amount_in, path=p)
            batcher.add(self.quoter.functions.quoteExactInput(q.route.encoded, amount_in))
            inputs.append(q)
        return batcher, inputs

    def prepare_quotes(self, quote_inputs: List[QuoteInput], responses):
        if len(responses) != len(quote_inputs): raise ValueError(f"Number of responses {len(responses)} does not match number of quote inputs {len(quote_inputs)}")
        quotes = []
        for i, r in enumerate(responses):
            if isinstance(r, Exception): continue
            else: quotes.append(Quote(input=quote_inputs[i], amount_out=r[0]))
        return quotes
    
    def get_paths_for_quote(self, from_token: Token, to_token: Token, pools: List[LiquidityPoolForSwap], exclude_tokens: List[str]) -> List[List[Tuple]]:
        exclude_tokens_set = set(map(lambda t: normalize_address(t), exclude_tokens))

        if from_token.token_address in exclude_tokens: exclude_tokens_set.remove(from_token.token_address)
        if to_token.token_address in exclude_tokens: exclude_tokens_set.remove(to_token.token_address)

        pairs = [Pair(p.token0_address, p.token1_address, p.lp) for p in pools]
        paths = find_all_paths(pairs, from_token.wrapped_token_address or from_token.token_address, to_token.wrapped_token_address or to_token.token_address)
        # filter out paths with excluded tokens
        return list(filter(lambda p: len(set(map(lambda t: t[0], p)) & exclude_tokens_set) == 0, paths))

    def get_pool_paginator(self, batch_size = 5) -> List[List[Tuple]]:
        limit, upper_bound = self.settings.pool_page_size, self.settings.pools_count_upper_bound
        return chunk(list(map(lambda x: (x, limit), list(range(0, upper_bound, limit)))), batch_size)
    
    def prepare_price_batcher(self, tokens: List[Token], batch: RequestBatcher):
        batches = chunk(tokens, self.settings.price_batch_size)
        for b in batches:
            batch.add(self.prices.functions.getManyRatesToEthWithCustomConnectors(
                list(map(lambda t: t.wrapped_token_address or t.token_address, b)),
                False, # use wrappers
                self.settings.connector_tokens_addrs,
                10 # threshold_filter
            ))
        return batch        

# %% ../src/chains.ipynb 8
class AsyncChain(CommonChain):
    web3: AsyncWeb3
    sugar: AsyncContract
    slipstream: AsyncContract
    router: AsyncContract
    prices: AsyncContract
    quoter: AsyncContract
    swapper: AsyncContract


    async def __aenter__(self):
        """Async context manager entry"""
        self._in_context = True
        self.web3 = AsyncWeb3(AsyncHTTPProvider(self.settings.rpc_uri))
        self.sugar = self.web3.eth.contract(address=self.settings.sugar_contract_addr, abi=get_abi("sugar"))
        self.sugar_rewards = self.web3.eth.contract(address=self.settings.sugar_rewards_contract_addr, abi=get_abi("sugar_rewards"))
        self.slipstream = self.web3.eth.contract(address=self.settings.slipstream_contract_addr, abi=get_abi("slipstream"))
        self.prices = self.web3.eth.contract(address=self.settings.price_oracle_contract_addr, abi=get_abi("price_oracle"))
        self.router = self.web3.eth.contract(address=self.settings.router_contract_addr, abi=get_abi("router"))
        self.quoter = self.web3.eth.contract(address=self.settings.quoter_contract_addr, abi=get_abi("quoter"))
        self.swapper = self.web3.eth.contract(address=self.settings.swapper_contract_addr, abi=get_abi("swapper"))

        # set up caching for price oracle
        self._get_prices = alru_cache(ttl=self.settings.pricing_cache_timeout_seconds)(self._get_prices)

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit"""
        self._in_context = False
        await self.web3.provider.disconnect()
        return None

    async def apaginate(self, f: Callable):
        async def process_batch(batch: List[Tuple]):
            async with self.web3.batch_requests() as batcher:
                for offset, limit in batch: batcher.add(f(limit, offset))
                return sum(await batcher.async_execute(), [])
        return sum(await asyncio.gather(*[process_batch(batch) for batch in self.get_pool_paginator()]), [])

    @require_async_context
    @alru_cache(maxsize=None)
    async def get_all_tokens(self, listed_only: bool = False) -> List[Token]:
        def get_tokens(limit, offset): return self.sugar.functions.tokens(limit, offset, ADDRESS_ZERO, [])
        return self.prepare_tokens(await self.apaginate(get_tokens), listed_only)

    async def _get_prices(self, tokens: Tuple[Token]):
        async with self.web3.batch_requests() as batch:
            batch = self.prepare_price_batcher(tokens=list(tokens), batch=batch)
            return sum(await batch.async_execute(), [])

    @require_async_context
    async def get_prices(self, tokens: List[Token]) -> List[Price]:
        """Get prices for tokens in target stable token"""
        return self.prepare_prices(tokens, await self._get_prices(tuple(tokens)))

    @alru_cache(maxsize=None)
    async def get_raw_pools(self, for_swaps: bool):
        return await self.apaginate(self.sugar.functions.forSwaps if for_swaps else self.sugar.functions.all)
    
    @require_async_context
    async def get_pools(self, for_swaps: bool = False) -> List[LiquidityPool]:
        pools = await self.get_raw_pools(for_swaps)
        if not for_swaps:
            tokens = await self.get_all_tokens()
            return self.prepare_pools(pools, tokens, await self.get_prices(tokens))
        else: return self.prepare_pools_for_swap(pools)
    
    @require_async_context
    @alru_cache(maxsize=None)
    async def get_pool_by_address(self, address: str) -> Optional[LiquidityPool]:
        try:
            p = await self.sugar.functions.byAddress(address).call()
        except: return None
        tokens = await self.get_all_tokens(listed_only=False)
        return self.prepare_pools([p], tokens, await self.get_prices(tokens))[0]

    @require_async_context
    @alru_cache(maxsize=None)
    async def get_pool_epochs(self, lp: str, offset: int = 0, limit: int = 10) -> List[LiquidityPoolEpoch]:
        tokens, pools = await self.get_all_tokens(listed_only=False), await self.get_pools()
        prices = await self.get_prices(tokens)
        r = await self.sugar_rewards.functions.epochsByAddress(limit, offset, normalize_address(lp)).call()
        return self.prepare_pool_epochs(r, pools, tokens, prices)

    @require_async_context
    @alru_cache(maxsize=None)
    async def get_latest_pool_epochs(self) -> List[LiquidityPoolEpoch]:
        tokens, pools = await self.get_all_tokens(listed_only=False), await self.get_pools()
        prices = await self.get_prices(tokens)
        return self.prepare_pool_epochs(await self.apaginate(self.sugar_rewards.functions.epochsLatest), pools, tokens, prices)
    
    @require_async_context
    async def get_pools_for_swaps(self) -> List[LiquidityPoolForSwap]: return await self.get_pools(for_swaps=True)

    @require_async_context
    async def _get_quotes_for_paths(self, from_token: Token, to_token: Token, amount_in: int, pools: List[LiquidityPoolForSwap], paths: List[List[Tuple]]) -> List[Optional[Quote]]:
        path_pools = self.paths_to_pools(pools, paths)
        async with self.web3.batch_requests() as batch:
            batch, inputs = self.prepare_quote_batch(from_token, to_token, batch, path_pools, amount_in, paths)
            return self.prepare_quotes(inputs, await batch.async_execute())

    @require_async_context
    async def get_quote(self, from_token: Token, to_token: Token, amount: float, filter_quotes: Optional[Callable[[Quote], bool]] = None) -> Optional[Quote]:
        amount_in = float_to_uint256(amount, decimals=from_token.decimals)
        pools = self.filter_pools_for_swap(from_token=from_token, to_token=to_token, pools=await self.get_pools_for_swaps())
        paths = self.get_paths_for_quote(from_token, to_token, pools, self.settings.excluded_tokens_addrs)
        quotes = sum(await asyncio.gather(*[self._get_quotes_for_paths(from_token, to_token, amount_in, pools, paths) for paths in chunk(paths, 500)]), [])
        quotes = list(filter(lambda q: q is not None, quotes))
        if filter_quotes is not None: quotes = list(filter(filter_quotes, quotes))
        return max(quotes, key=lambda q: q.amount_out) if len(quotes) > 0 else None
    
    @require_async_context
    async def swap(self, from_token: Token, to_token: Token, amount: float, slippage: Optional[float] = None):
        q = await self.get_quote(from_token, to_token, amount)
        if not q: raise ValueError("No quotes found")
        return await self.swap_from_quote(q, slippage=slippage)
        
    @require_async_context
    async def swap_from_quote(self, quote: Quote, slippage: Optional[float] = None):
        swapper_contract_addr, from_token = self.settings.swapper_contract_addr, quote.from_token
        planner = setup_planner(quote=quote, slippage=slippage if slippage is not None else self.settings.swap_slippage, account=self.account.address, router_address=swapper_contract_addr)
        await self.set_token_allowance(from_token, swapper_contract_addr, quote.input.amount_in)
        value = quote.input.amount_in if from_token.wrapped_token_address else 0
        return await self.sign_and_send_tx(self.swapper.functions.execute(*[planner.commands, planner.inputs]), value=value)

    @require_async_context
    async def set_token_allowance(self, token: Token, addr: str, amount: int):
        token_contract = self.prepare_set_token_allowance_contract(token, self.web3.eth.contract)
        return await self.sign_and_send_tx(token_contract.functions.approve(addr, amount))

    @require_async_context
    async def check_token_allowance(self, token: Token, addr: str) -> int:
        ERC20_ABI = [{
            "name": "allowance",
            "type": "function",
            "constant": True,
            "inputs": [{"name": "owner", "type": "address"}, {"name": "spender", "type": "address"}],
            "outputs": [{"name": "", "type": "uint256"}]
        }]
        token_contract = self.web3.eth.contract(address=token.wrapped_token_address or token.token_address, abi=ERC20_ABI)
        return await token_contract.functions.allowance(self.account.address, addr).call()

    @require_async_context
    async def sign_and_send_tx(self, tx, value: int = 0, wait: bool = True):
        spender = self.account.address
        tx = await tx.build_transaction({ 'from': spender, 'value': value, 'nonce': await self.web3.eth.get_transaction_count(spender) })
        signed_tx = self.account.sign_transaction(tx)
        tx_hash = await self.web3.eth.send_raw_transaction(signed_tx.raw_transaction)
        return await self.web3.eth.wait_for_transaction_receipt(tx_hash) if wait else tx_hash
    
    @require_async_context
    async def deposit(self, deposit: Deposit, delay_in_minutes: float = 30, slippage: float = 0.01):
        amount_token0, pool, router_contract_addr = deposit.amount_token0, deposit.pool, self.settings.router_contract_addr
        print(f"gonna deposit {amount_token0} {pool.token0.symbol} into {pool.symbol} from {self.account.address}")
        [token0_amount, token1_amount, _] = await self.router.functions.quoteAddLiquidity(
            pool.token0.token_address,
            pool.token1.token_address,
            pool.is_stable,
            pool.factory,
            float_to_uint256(amount_token0, pool.token0.decimals),
            MAX_UINT256
        ).call()
        print(f"Quote: {pool.token0.symbol} {token0_amount / 10 ** pool.token0.decimals} -> {pool.token1.symbol} {token1_amount / 10 ** pool.token1.decimals}")

        # set up allowance for both tokens
        print(f"setting up allowance for {pool.token0.symbol}")
        await self.set_token_allowance(pool.token0, router_contract_addr, token0_amount)

        print(f"setting up allowance for {pool.token1.symbol}")
        await self.set_token_allowance(pool.token1, router_contract_addr, token1_amount)

        # check allowances
        token0_allowance = await self.check_token_allowance(pool.token0, router_contract_addr)
        token1_allowance = await self.check_token_allowance(pool.token1, router_contract_addr)

        print(f"allowances: {token0_allowance}, {token1_allowance}")

        # adding liquidity

        # if token 0 is native, use addLiquidityETH instead of standard addLiquidity
        if pool.token0.token_address == self.settings.wrapped_native_token_addr:
            params = [
                pool.token1.token_address,
                pool.is_stable,
                token1_amount,
                apply_slippage(token1_amount, slippage),
                apply_slippage(token0_amount, slippage),
                self.account.address,
                get_future_timestamp(delay_in_minutes)
            ]
            print(f"adding liquidity with params: {params}")
            return await self.sign_and_send_tx(self.router.functions.addLiquidityETH(*params), value=token0_amount)
        
        # token 1 is native, use addLiquidityETH instead of standard addLiquidity
        if pool.token1.token_address == self.settings.wrapped_native_token_addr:
            params = [
                pool.token0.token_address,
                pool.is_stable,
                token0_amount,
                apply_slippage(token0_amount, slippage),
                apply_slippage(token1_amount, slippage),
                self.account.address,
                get_future_timestamp(delay_in_minutes)
            ]
            print(f"adding liquidity with params: {params}")
            return await self.sign_and_send_tx(self.router.functions.addLiquidityETH(*params), value=token1_amount)

        params = [
            pool.token0.token_address,
            pool.token1.token_address,
            pool.is_stable,
            token0_amount,
            token1_amount,
            apply_slippage(token0_amount, slippage),
            apply_slippage(token1_amount, slippage),
            self.account.address,
            get_future_timestamp(delay_in_minutes)
        ]

        print(f"adding liquidity with params: {params}")

        return await self.sign_and_send_tx(self.router.functions.addLiquidity(*params))

# %% ../src/chains.ipynb 10
class Chain(CommonChain):
    web3: Web3
    sugar: Contract
    router: Contract
    slipstream: Contract
    prices: Contract
    quoter: Contract
    swapper: Contract

    def __enter__(self):
        """Sync context manager entry"""
        self._in_context = True
        self.web3 = Web3(HTTPProvider(self.settings.rpc_uri))
        self.sugar = self.web3.eth.contract(address=self.settings.sugar_contract_addr, abi=get_abi("sugar"))
        self.sugar_rewards = self.web3.eth.contract(address=self.settings.sugar_rewards_contract_addr, abi=get_abi("sugar_rewards"))
        self.slipstream = self.web3.eth.contract(address=self.settings.slipstream_contract_addr, abi=get_abi("slipstream"))
        self.prices = self.web3.eth.contract(address=self.settings.price_oracle_contract_addr, abi=get_abi("price_oracle"))
        self.router = self.web3.eth.contract(address=self.settings.router_contract_addr, abi=get_abi("router"))
        self.quoter = self.web3.eth.contract(address=self.settings.quoter_contract_addr, abi=get_abi("quoter"))
        self.swapper = self.web3.eth.contract(address=self.settings.swapper_contract_addr, abi=get_abi("swapper"))

        # set up caching for price oracle
        self._get_prices = cached(TTLCache(ttl=self.settings.pricing_cache_timeout_seconds, maxsize=self.settings.price_batch_size * 10))(self._get_prices)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Sync context manager exit"""
        self._in_context = False
        return None
    
    def paginate(self, f: Callable):
        results, batches = [], self.get_pool_paginator()

        def process_batch(batch: List[Tuple]):
            with self.web3.batch_requests() as batcher:
                for offset, limit in batch: batcher.add(f(limit, offset))
                return sum(batcher.execute(), [])

        
        with ThreadPoolExecutor(max_workers=self.settings.threading_max_workers) as executor:
            future_to_batch = {
                executor.submit(process_batch, batch): batch
                for batch in batches
            }
            for future in as_completed(future_to_batch):
                try: results.extend(future.result())
                except Exception as e:
                    print(f"Error processing path chunk: {e}")
                    continue

        return results


    @require_context
    def sign_and_send_tx(self, tx, value: int = 0, wait: bool = True):
        spender = self.account.address
        tx = tx.build_transaction({ 'from': spender, 'value': value, 'nonce': self.web3.eth.get_transaction_count(spender) })
        signed_tx = self.account.sign_transaction(tx)
        tx_hash = self.web3.eth.send_raw_transaction(signed_tx.raw_transaction)
        return self.web3.eth.wait_for_transaction_receipt(tx_hash) if wait else tx_hash
    
    @require_context
    def set_token_allowance(self, token: Token, addr: str, amount: int):
        token_contract = self.prepare_set_token_allowance_contract(token, self.web3.eth.contract)
        return self.sign_and_send_tx(token_contract.functions.approve(addr, amount))

    @require_context
    @lru_cache(maxsize=None)
    def get_all_tokens(self, listed_only: bool = False) -> List[Token]:
        def get_tokens(limit, offset): return self.sugar.functions.tokens(limit, offset, ADDRESS_ZERO, [])
        return self.prepare_tokens(self.paginate(get_tokens), listed_only)
    
    def _get_prices(self, tokens: Tuple[Token]) -> List[int]:
        # token_address => normalized rate
        with self.web3.batch_requests() as batch:
            batch = self.prepare_price_batcher(tokens=list(tokens), batch=batch)
            return sum(batch.execute(), [])

    @require_context
    def get_prices(self, tokens: List[Token]) -> List[Price]:
        """Get prices for tokens in target stable token"""
        return self.prepare_prices(tokens, self._get_prices(tuple(tokens)))
    
    @lru_cache(maxsize=None)
    def get_raw_pools(self, for_swaps: bool):
        return self.paginate(self.sugar.functions.forSwaps if for_swaps else self.sugar.functions.all)

    @require_context
    def get_pools(self, for_swaps: bool = False) -> List[LiquidityPool]:
        pools = self.get_raw_pools(for_swaps)
        if not for_swaps:
            tokens = self.get_all_tokens(listed_only=False)
            return self.prepare_pools(pools, tokens, self.get_prices(tokens))
        else: return self.prepare_pools_for_swap(pools)

    @require_context
    @lru_cache(maxsize=None)
    def get_pool_by_address(self, address: str) -> Optional[LiquidityPool]:
        try:
            p = self.sugar.functions.byAddress(address).call()
        except: return None
        tokens = self.get_all_tokens(listed_only=False)
        return self.prepare_pools([p], tokens, self.get_prices(tokens))[0]
    
    @require_context
    def get_pools_for_swaps(self) -> List[LiquidityPoolForSwap]: return self.get_pools(for_swaps=True)

    @require_context
    @lru_cache(maxsize=None)
    def get_pool_epochs(self, lp: str, offset: int = 0, limit: int = 10) -> List[LiquidityPoolEpoch]:
        tokens, pools = self.get_all_tokens(listed_only=False), self.get_pools()
        prices = self.get_prices(tokens)
        r = self.sugar_rewards.functions.epochsByAddress(limit, offset, normalize_address(lp)).call()
        return self.prepare_pool_epochs(r, pools, tokens, prices)

    @require_context
    @lru_cache(maxsize=None)
    def get_latest_pool_epochs(self) -> List[LiquidityPoolEpoch]:
        tokens, pools = self.get_all_tokens(listed_only=False), self.get_pools()
        prices = self.get_prices(tokens)
        return self.prepare_pool_epochs(self.paginate(self.sugar_rewards.functions.epochsLatest), pools, tokens, prices)

    @require_context
    def _get_quotes_for_paths(self, from_token: Token, to_token: Token, amount_in: int, pools: List[LiquidityPoolForSwap], paths: List[List[Tuple]]) -> List[Optional[Quote]]:
        path_pools = self.paths_to_pools(pools, paths)
        with self.web3.batch_requests() as batch:
            batch, inputs = self.prepare_quote_batch(from_token, to_token, batch, path_pools, amount_in, paths)
            return self.prepare_quotes(inputs, batch.execute())
    
    @require_context
    def get_quote(self, from_token: Token, to_token: Token, amount: float, filter_quotes: Optional[Callable[[Quote], bool]] = None) -> Optional[Quote]:
        amount_in = float_to_uint256(amount, decimals=from_token.decimals)
        pools = self.filter_pools_for_swap(from_token=from_token, to_token=to_token, pools=self.get_pools_for_swaps())
        paths = self.get_paths_for_quote(from_token, to_token, pools, self.settings.excluded_tokens_addrs)
        path_chunks = list(chunk(paths, 500))

        def get_quotes_for_chunk(paths_chunk):
            return self._get_quotes_for_paths(from_token, to_token, amount_in, pools, paths_chunk)
        
        all_quotes = []
        
        with ThreadPoolExecutor(max_workers=self.settings.threading_max_workers) as executor:
            # Submit all chunk processing tasks
            future_to_chunk = {
                executor.submit(get_quotes_for_chunk, chunk_paths): chunk_paths 
                for chunk_paths in path_chunks
            }
            
            # Collect results as they complete
            for future in as_completed(future_to_chunk):
                try:
                    all_quotes.extend(future.result())
                except Exception as e:
                    print(f"Error processing path chunk: {e}")
                    continue

        # Filter out None quotes
        all_quotes = [q for q in all_quotes if q is not None]
    
        if filter_quotes is not None:  all_quotes = list(filter(filter_quotes, all_quotes))

        return max(all_quotes, key=lambda q: q.amount_out) if len(all_quotes) > 0 else None
    
    @require_context
    def swap(self, from_token: Token, to_token: Token, amount: float, slippage: Optional[float] = None):
        q = self.get_quote(from_token, to_token, amount)
        if not q: raise ValueError("No quotes found")
        return self.swap_from_quote(q, slippage=slippage)
        
    @require_context
    def swap_from_quote(self, quote: Quote, slippage: Optional[float] = None):
        swapper_contract_addr, from_token = self.settings.swapper_contract_addr, quote.from_token
        planner = setup_planner(quote=quote, slippage=slippage if slippage is not None else self.settings.swap_slippage, account=self.account.address, router_address=swapper_contract_addr)
        self.set_token_allowance(from_token, swapper_contract_addr, quote.input.amount_in)
        value = quote.input.amount_in if from_token.wrapped_token_address else 0
        return self.sign_and_send_tx(self.swapper.functions.execute(*[planner.commands, planner.inputs]), value=value)

# %% ../src/chains.ipynb 12
class OPChainCommon():
    usdc: Token = Token(token_address='0x0b2C639c533813f4Aa9D7837CAf62653d097Ff85', symbol='USDC', decimals=6, listed=True, wrapped_token_address=None)
    velo: Token = Token(token_address='0x9560e827aF36c94D2Ac33a39bCE1Fe78631088Db', symbol='VELO', decimals=18, listed=True, wrapped_token_address=None)
    eth: Token = Token(token_address='ETH', symbol='ETH', decimals=18, listed=True, wrapped_token_address='0x4200000000000000000000000000000000000006')
    
class AsyncOPChain(AsyncChain, OPChainCommon):
    def __init__(self, **kwargs): super().__init__(make_op_chain_settings(**kwargs), **kwargs)

class OPChain(Chain, OPChainCommon):
    def __init__(self, **kwargs): super().__init__(make_op_chain_settings(**kwargs), **kwargs)


# %% ../src/chains.ipynb 14
class BaseChainCommon():
    usdc: Token = Token(token_address='0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913', symbol='USDC', decimals=6, listed=True, wrapped_token_address=None)
    aero: Token = Token(token_address='0x940181a94A35A4569E4529A3CDfB74e38FD98631', symbol='AERO', decimals=18, listed=True, wrapped_token_address=None)
    eth: Token = Token(token_address='ETH', symbol='ETH', decimals=18, listed=True, wrapped_token_address='0x4200000000000000000000000000000000000006')
    
class AsyncBaseChain(AsyncChain, BaseChainCommon):
    def __init__(self, **kwargs): super().__init__(make_base_chain_settings(**kwargs), **kwargs)

class BaseChain(Chain, BaseChainCommon):
    def __init__(self, **kwargs): super().__init__(make_base_chain_settings(**kwargs), **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/config.ipynb.

# %% auto 0
__all__ = ['base_default_settings', 'GetEnv', 'ChainSettings', 'validate_settings', 'get_env', 'make_settings',
           'make_op_chain_settings', 'make_base_chain_settings']

# %% ../src/config.ipynb 3
import os
from dataclasses import dataclass, make_dataclass, fields
from .helpers import normalize_address
from typing import List, Callable, Any, Dict

# %% ../src/config.ipynb 4
base_default_settings = {
  "price_batch_size": int(os.getenv("SUGAR_PRICE_BATCH_SIZE","40")),
  "price_threshold_filter": int(os.getenv("SUGAR_PRICE_THRESHOLD_FILTER","10")),
  "pagination_limit": int(os.getenv("SUGAR_PAGINATION_LIMIT","2000")),
  "pool_page_size": int(os.getenv("SUGAR_POOL_PAGE_SIZE","500")),
  # XX: dealing with Schrödinger's paginator this is likely to be ignored in the future with new sugar helpers
  "pools_count_upper_bound": 2500,
  "native_token_symbol": "ETH",
  "native_token_decimals": 18,
  "swap_slippage": 0.01,
  "pricing_cache_timeout_seconds": 5,
  "threading_max_workers": 5
}


@dataclass
class ChainSettings:
    # chain IDs come from: https://chainlist.org/
    chain_id: str
    chain_name: str
    wrapped_native_token_addr: str
    rpc_uri: str
    sugar_contract_addr: str
    sugar_rewards_contract_addr: str
    # slipstream operates on concentrated liquidity (CL) pools
    slipstream_contract_addr: str
    # Non-Fungible Position Manager for CL pools
    nfpm_contract_addr: str
    price_oracle_contract_addr: str
    router_contract_addr: str
    quoter_contract_addr: str
    # aka Universal Router
    swapper_contract_addr: str
    token_addr: str
    stable_token_addr: str
    connector_tokens_addrs: List[str]
    # tokens to exclude from quote search graph
    excluded_tokens_addrs: List[str]
    # default swap slippage in % (0.0-1.0)
    swap_slippage: float
    price_batch_size: int
    price_threshold_filter: int
    pagination_limit: int
    pools_count_upper_bound: int
    pool_page_size: int
    native_token_symbol: str
    native_token_decimals: int
    # how often to check for new prices
    pricing_cache_timeout_seconds: int
    # how many max workers to use for threading in sync methods
    threading_max_workers: int

    def __str__(self):
        # go over all attributes of self
        lines = ["🍭 Chain settings:","----------------"]
        attributes = [attr for attr in dir(self) if not attr.startswith('_')]
        for attr in attributes:
            # Skip methods
            if callable(getattr(self, attr)): continue
            value = getattr(self, attr)
            if isinstance(value, list):
                value = "\n" + "\n".join([f"{' ' * 4 }- {v}" for v in value])
            lines.append(f"{attr}: {value}")
        return "\n".join(lines)
    
    def __repr__(self): return str(self)

def validate_settings(settings: ChainSettings) -> ChainSettings:
    # TODO: this should actually validate stuff, duh
    floats = ["swap_slippage"]
    ints = ["price_batch_size", "price_threshold_filter", "pagination_limit", "pool_page_size", "native_token_decimals",
            "pricing_cache_timeout_seconds", "pools_count_upper_bound", "threading_max_workers"]
    for k in floats: setattr(settings, k, float(getattr(settings, k)))
    for k in ints: setattr(settings, k, int(getattr(settings, k)))
    return settings

def get_env(key: str, default: Any) -> Any: return os.getenv(key, default)

GetEnv = Callable[[str, Any], Any]

def make_settings(chain_id: str, chain_name: str, chain_settings: Dict[str, Any], get_env: GetEnv, **kwargs) -> ChainSettings:
    d = { **base_default_settings, **chain_settings, **{ "chain_id": chain_id, "chain_name": chain_name } }

    for k,v in d.items():
        if k in ["chain_id", "chain_name"]: continue
        # look for env vars (i.e stable_token_addr for chain_id="10" should be SUGAR_STABLE_TOKEN_ADDR_10)
        d[k] = get_env(f"SUGAR_{k.upper()}_{chain_id}", v)

    # keywords override env vars
    d = d | kwargs

    for k,v in d.items():
        # anything that ends in "_addr", should be normalized
        if k.endswith("_addr"): d[k] = normalize_address(d[k])
        # anything that ends in "_addrs", should be a list of normalized addresses
        if k.endswith("_addrs"): d[k] = list(map(lambda a: normalize_address(a), d[k].split(",")))

    # we only want fields that are in the ChainSettings dataclass
    d =  {k: v for k, v in d.items() if k in [field.name for field in fields(ChainSettings)]}

    return validate_settings(make_dataclass(ChainSettings.__name__, ((k, type(v)) for k, v in d.items()))(**d)) 

# %% ../src/config.ipynb 6
def make_op_chain_settings(get_env: GetEnv = get_env, **kwargs) -> ChainSettings:
    d = {
        "rpc_uri": "https://optimism-mainnet.wallet.coinbase.com",
        "wrapped_native_token_addr": "0x4200000000000000000000000000000000000006",
        "sugar_contract_addr": "0xA64db2D254f07977609def75c3A7db3eDc72EE1D",
        "sugar_rewards_contract_addr": "0x62CCFB2496f49A80B0184AD720379B529E9152fB",
        "slipstream_contract_addr": "0xD45624bf2CB9f65ecbdF3067d21992b099b56202",
        "nfpm_contract_addr": "0x416b433906b1B72FA758e166e239c43d68dC6F29",
        "price_oracle_contract_addr": "0x59114D308C6DE4A84F5F8cD80485a5481047b99f",
        "router_contract_addr": "0xa062aE8A9c5e11aaA026fc2670B0D65cCc8B2858",
        "quoter_contract_addr": "0xFF79ec912bA114FD7989b9A2b90C65f0c1b44722",
        "swapper_contract_addr": "0x4bF3E32de155359D1D75e8B474b66848221142fc",
        "token_addr": "0x9560e827aF36c94D2Ac33a39bCE1Fe78631088Db",
        "stable_token_addr": "0x7f5c764cbc14f9669b88837ca1490cca17c31607",
        "connector_tokens_addrs": "0x9560e827aF36c94D2Ac33a39bCE1Fe78631088Db,0x4200000000000000000000000000000000000042,0x4200000000000000000000000000000000000006,0x9bcef72be871e61ed4fbbc7630889bee758eb81d,0x2e3d870790dc77a83dd1d18184acc7439a53f475,0x8c6f28f2f1a3c87f0f938b96d27520d9751ec8d9,0x1f32b1c2345538c0c6f582fcb022739c4a194ebb,0xbfd291da8a403daaf7e5e9dc1ec0aceacd4848b9,0xc3864f98f2a61a7caeb95b039d031b4e2f55e0e9,0x9485aca5bbbe1667ad97c7fe7c4531a624c8b1ed,0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1,0x73cb180bf0521828d8849bc8cf2b920918e23032,0x6806411765af15bddd26f8f544a34cc40cb9838b,0x6c2f7b6110a37b3b0fbdd811876be368df02e8b0,0xc5b001dc33727f8f26880b184090d3e252470d45,0x6c84a8f1c29108f47a79964b5fe888d4f4d0de40,0xc40f949f8a4e094d1b49a23ea9241d289b7b2819,0x94b008aa00579c1307b0ef2c499ad98a8ce58e58,0x0b2c639c533813f4aa9d7837caf62653d097ff85",
        "excluded_tokens_addrs": "0x74ccbe53f77b08632ce0cb91d3a545bf6b8e0979,0x139283255069ea5deef6170699aaef7139526f1f,0x88a89866439f4c2830986b79cbe6f69d1bd548bb,0x8901cb2e82cc95c01e42206f8d1f417fe53e7af0"
    } 
    return make_settings("10", "OP", chain_settings=d, get_env=get_env, **kwargs)

# %% ../src/config.ipynb 8
def make_base_chain_settings(get_env: GetEnv = get_env, **kwargs) -> ChainSettings:
    d = {
        "rpc_uri": "https://mainnet.base.org",
        "pools_count_upper_bound": 9000,
        "wrapped_native_token_addr": "0x4200000000000000000000000000000000000006",
        "sugar_contract_addr": "0x6F8Ea68A1a66e49e16A470bcF6fe2A3a7b94CDe9",
        "sugar_rewards_contract_addr": "0xA44600F4DBA6683d8BD99270B1A6a143fB9F1C3B",
        "slipstream_contract_addr": "0x0AD09A66af0154a84e86F761313d02d0abB6edd5",
        "nfpm_contract_addr": "0x827922686190790b37229fd06084350E74485b72",
        "price_oracle_contract_addr": "0x3B06c787711ecb5624cE65AC8F26cde10831eb0C",
        "router_contract_addr": "0xcF77a3Ba9A5CA399B7c97c74d54e5b1Beb874E43",
        "quoter_contract_addr": "0x0A5aA5D3a4d28014f967Bf0f29EAA3FF9807D5c6",
        "swapper_contract_addr": "0x6Cb442acF35158D5eDa88fe602221b67B400Be3E",
        "token_addr": "0x9560e827aF36c94D2Ac33a39bCE1Fe78631088Db",
        "stable_token_addr": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
        "connector_tokens_addrs": "0x833589fcd6edb6e08f4c7c32d4f71b54bda02913,0x940181a94A35A4569E4529A3CDfB74e38FD98631,0x50c5725949A6F0c72E6C4a641F24049A917DB0Cb,0x4621b7a9c75199271f773ebd9a499dbd165c3191,0x4200000000000000000000000000000000000006,0xb79dd08ea68a908a97220c76d19a6aa9cbde4376,0xf7a0dd3317535ec4f4d29adf9d620b3d8d5d5069,0xcfa3ef56d303ae4faaba0592388f19d7c3399fb4,0xcb327b99ff831bf8223cced12b1338ff3aa322ff,0x2ae3f1ec7f1f5012cfeab0185bfc7aa3cf0dec22,0xc1cba3fcea344f92d9239c08c0568f6f2f0ee452,0x60a3e35cc302bfa44cb288bc5a4f316fdb1adb42,0xd9aaec86b65d86f6a7b5b1b0c42ffa531710b6ca,0xcbB7C0000aB88B473b1f5aFd9ef808440eed33Bf",
        "excluded_tokens_addrs": "0x74ccbe53f77b08632ce0cb91d3a545bf6b8e0979,0x8901cb2e82cc95c01e42206f8d1f417fe53e7af0,0x9cbd543f1b1166b2df36b68eb6bb1dce24e6abdf,0x025f99977db78317a4eba606998258b502bb256f,0xd260115030b9fb6849da169a01ed80b6496d1e99,0x608d5401d377228e465ba6113517dcf9bd1f95ca,0xd260115030b9fB6849da169a01ed80b6496d1e99,0x728cDA34D732a87fD6429129e23D4742d9Ff0064,0x728cda34d732a87fd6429129e23d4742d9ff0064,0xAC1Bd2486aAf3B5C0fc3Fd868558b082a531B2B4,0x0f929C29dcE303F96b1d4104505F2e60eE795caC,0x47E78d664E6c339693e8638B7A7D9543AbCc99D4,0xFF0C532FDB8Cd566Ae169C1CB157ff2Bdc83E105,0x373504da48418c67e6fcd071f33cb0b3b47613c7,0x0f929c29dce303f96b1d4104505f2e60ee795cac,0x628c5Ba9B775DACEcd14E237130c537f497d1CC7"
    }
    return make_settings("8453", "Base", chain_settings=d, get_env=get_env, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/abi.ipynb.

# %% auto 0
__all__ = ['abis_dir', 'download_contract_abi', 'get_abi']

# %% ../src/abi.ipynb 3
import requests, os, json

# %% ../src/abi.ipynb 4
abis_dir="abis"

def download_contract_abi(name, address, abis_dir=os.path.join("sugar", abis_dir), etherscan_api_url="https://api-optimistic.etherscan.io/api"):
    api_key = os.getenv("ETHERSCAN_API_KEY")
    if not api_key: raise Exception("ETHERSCAN_API_KEY not set in environment variables")
    response = requests.get(etherscan_api_url, params={ "chainid": 1, "module": "contract", "action": "getabi", "address": address, "apikey": api_key })
    if response.status_code != 200: raise Exception(f"API request failed with status code {response.status_code}")
    response_data = response.json()
    if response_data.get("message") != "OK": raise Exception(f"API request failed with message: {response_data.get('message')}")
    os.makedirs(abis_dir, exist_ok=True)
    with open(os.path.join(abis_dir, f"{name}.json"), "w") as file:
        json.dump(json.loads(response_data.get("result")), file, indent=4)

def get_abi(name):
    dir, path = None, os.path.abspath(__file__)
    dir_path = os.path.dirname(path)
    abis_locations = [
        abis_dir,
        os.path.join(dir_path, abis_dir),
        os.path.join("../", abis_dir),
        os.path.join("../sugar", abis_dir),
    ]

    for d in abis_locations:
        if os.path.exists(d):
            dir = d
            break

    if not dir: raise Exception("Cannot find abis directory. Did you run download_contract_abi?")
        
    with open(os.path.join(dir, f"{name}.json"), "r") as file:
        return file.read()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/helpers.ipynb.

# %% auto 0
__all__ = ['ADDRESS_ZERO', 'MAX_UINT256', 'normalize_address', 'chunk', 'amount_to_k_string', 'format_currency',
           'format_percentage', 'amount_to_m_string', 'float_to_uint256', 'get_future_timestamp', 'apply_slippage',
           'Pair', 'find_all_paths', 'Timer', 'time_it', 'atime_it']

# %% ../src/helpers.ipynb 2
from web3 import Web3, constants
from typing import List, Tuple, Optional, Callable
from decimal import Decimal
from datetime import datetime, timedelta
from dataclasses import dataclass
import networkx as nx
import math, time, asyncio
from contextlib import contextmanager, asynccontextmanager

# %% ../src/helpers.ipynb 3
def normalize_address(address: str) -> str: return Web3.to_checksum_address(address.lower())

ADDRESS_ZERO = constants.ADDRESS_ZERO
MAX_UINT256 = Web3.to_int(hexstr='0x' + 'f' * 64)

def chunk(list_to_chunk: List, n: int):
    for i in range(0, len(list_to_chunk), n):
        yield list_to_chunk[i : i + n]


def amount_to_k_string(amount: float) -> str:
    """Turns 2000 to "2K" """
    return f"{round(amount/1000, 2)}K"


def format_currency(value: float, symbol: str = "$", prefix: bool = True) -> str:
    v = "{:0,.2f}".format(value)
    return f"{symbol}{v}" if prefix else f"{v} {symbol}"


def format_percentage(value: float) -> str:
    return "{:0,.2f} %".format(value)


def amount_to_m_string(amount: float) -> str:
    """Turns 2000000 to "2M" """
    return f"{round(amount/1000000, 2)}M"

# %% ../src/helpers.ipynb 4
def float_to_uint256(amount: float, decimals: int = 18) -> int:
    """Convert float to uint256 with decimal scaling"""
    # Convert float to Decimal for precision
    amount_decimal = Decimal(str(amount))
    # Scale by decimals
    scaled_amount = amount_decimal * Decimal(10 ** decimals)
    # Convert to integer
    return int(scaled_amount)

# %% ../src/helpers.ipynb 6
def get_future_timestamp(deadline_minutes: float) -> int:
    """Convert minutes from now to future unix timestamp"""
    future_time = datetime.now() + timedelta(minutes=deadline_minutes)
    return int(future_time.timestamp())

# %% ../src/helpers.ipynb 8
def apply_slippage(amount: int, slippage: float) -> int:
    if slippage < 0 or slippage > 1: raise ValueError("Slippage must be between 0 and 1")
    return int(math.ceil(amount * (1 - slippage)))

# %% ../src/helpers.ipynb 12
# Claude 3.7 sonnet made this

@dataclass
class Pair: token0: str; token1: str; pool: str

def find_all_paths(pairs: List[Pair], start_token: str, end_token: str, cutoff=3) -> List[List[Tuple]]:
    # MultiGraph required to support parallel edges
    # same tokens can be present in different pools, hence parallel edges
    # specific pool identifier is stored inside edge attribute
    G, complete_paths = nx.MultiGraph(), []
    for pair in pairs: G.add_edge(pair.token0, pair.token1, pool=pair.pool)
    node_paths =  [p for p in nx.all_simple_paths(G, source=start_token, target=end_token, cutoff=cutoff)]
    for path in node_paths:
        edge_path = []
        # For each consecutive pair of nodes in the path
        for i in range(len(path) - 1):
            current = path[i]
            next_node = path[i + 1]
            
            # Get all edges between these nodes
            edges = G.get_edge_data(current, next_node)
            
            # There might be multiple edges (pools) between these nodes
            # Add all possible edges to create different complete paths
            current_paths = [] if not edge_path else edge_path.copy()
            new_edge_paths = []
            
            # If this is the first segment, initialize with empty path
            if not current_paths:
                current_paths = [[]]
                
            # For each possible edge between current and next_node
            for edge_key, edge_attrs in edges.items():
                pool = edge_attrs['pool']
                for current_path in current_paths:
                    # Create a new path that includes this edge
                    new_path = current_path + [(current, next_node, pool)]
                    new_edge_paths.append(new_path)
            
            edge_path = new_edge_paths
        
        # Add all possible edge paths to the complete paths
        complete_paths.extend(edge_path)
    
    # seen is a list of strings that look like this ["pool1-pool2-pool3", ...]
    uniques, seen = [], []

    for path in complete_paths:
        p = '-'.join(map(lambda x: x[2], path))
        if p not in seen:
            uniques.append(path)
            seen.append(p)

    # remove duplicates
    return uniques


# %% ../src/helpers.ipynb 15
# Claude 4 sonnet made this

class Timer:
    """Simple timer utility for measuring execution time"""
    
    def __init__(self, name: str = "Operation", precision: int = 4, callback: Optional[Callable] = None):
        self.name = name
        self.precision = precision
        self.callback = callback
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None
        self.elapsed: Optional[float] = None
    
    def __enter__(self):
        self.start_time = time.perf_counter()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = time.perf_counter()
        self.elapsed = self.end_time - self.start_time
        result = f"{self.name} took {self.elapsed:.{self.precision}f} seconds"
        
        if self.callback:
            self.callback(self.elapsed, result)
        else:
            print(result)
    
    async def __aenter__(self):
        self.start_time = time.perf_counter()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.end_time = time.perf_counter()
        self.elapsed = self.end_time - self.start_time
        result = f"{self.name} took {self.elapsed:.{self.precision}f} seconds"
        
        if self.callback:
            if asyncio.iscoroutinefunction(self.callback):
                await self.callback(self.elapsed, result)
            else:
                self.callback(self.elapsed, result)
        else:
            print(result)

@contextmanager
def time_it(name: str = "Operation", precision: int = 4, callback: Optional[Callable] = None):
    """Context manager for timing synchronous code execution"""
    timer = Timer(name, precision, callback)
    with timer:
        yield timer

@asynccontextmanager
async def atime_it(name: str = "Operation", precision: int = 4, callback: Optional[Callable] = None):
    """Async context manager for timing asynchronous code execution"""
    timer = Timer(name, precision, callback)
    async with timer:
        yield timer



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/quote.ipynb.

# %% auto 0
__all__ = ['QUOTER_STABLE_POOL_FILLER', 'QUOTER_VOLATILE_POOL_FILLER', 'PreparedRoute', 'pack_path', 'QuoteInput', 'Quote']

# %% ../src/quote.ipynb 3
from functools import reduce
from typing import List, Union, Tuple
from eth_abi.packed import encode_packed
from dataclasses import dataclass
from .token import Token
from .pool import LiquidityPoolForSwap

# %% ../src/quote.ipynb 4
# magic numbers
QUOTER_STABLE_POOL_FILLER, QUOTER_VOLATILE_POOL_FILLER = 2097152, 4194304

@dataclass
class PreparedRoute:
    types: List[str]; values: List[Union[str, int]]

    @property
    def encoded(self) -> bytes: return encode_packed(self.types, self.values)

def pack_path(path: List[Tuple[LiquidityPoolForSwap, bool]]) -> PreparedRoute:
    types, values = reduce(lambda s, pool: s + pool, [["address", "int24"] for i in range(len(path))], []) + ["address"], []
    for node in path:
        pool, reversed = node
        token0, token1 = pool.token0_address if not reversed else pool.token1_address, pool.token1_address if not reversed else pool.token0_address
        if pool.type > 0: filler = pool.type
        else: filler =  QUOTER_STABLE_POOL_FILLER if pool.is_stable else QUOTER_VOLATILE_POOL_FILLER
        if len(values) == 0: values = [token0, filler, token1]
        else: values += [filler, token1]
    return PreparedRoute(types=types, values=values)

@dataclass
class QuoteInput:
    from_token: Token; to_token: Token; path: List[Tuple[LiquidityPoolForSwap, bool]]; amount_in: int

    def to_tuple(self) -> tuple: return (self.from_token, self.to_token, tuple(self.path), self.amount_in, self.amount_out)

    @staticmethod
    def from_tuple(t: tuple) -> "QuoteInput":
        from_token, to_token, path, amount_in, amount_out = t
        return QuoteInput(from_token=from_token, to_token=to_token, path=list(path), amount_in=amount_in, amount_out=amount_out)
    
    @property
    def route(self) -> PreparedRoute: return pack_path(self.path)

@dataclass
class Quote:
    input: QuoteInput; amount_out: int

    @property
    def from_token(self) -> Token: return self.input.from_token
    @property
    def to_token(self) -> Token: return self.input.to_token
    @property
    def path(self) -> List[Tuple[LiquidityPoolForSwap, bool]]: return self.input.path
    @property
    def amount_in(self) -> int: return self.input.amount_in
